package org.netkernel.intray.transport;

import java.io.File;
import java.net.URI;
import java.util.Arrays;
import java.util.Comparator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.util.RegexRewriterRule;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.module.standard.endpoint.StandardTransportImpl;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

public class IntrayTransport extends StandardTransportImpl
{
	private IntrayProcessor mProc;
	
	@Override
	protected void postCommission(INKFRequestContext context) throws Exception
	{	mProc=new IntrayProcessor(context, this);
	}

	@Override
	protected void preDecommission(INKFRequestContext context) throws Exception
	{	if(mProc!=null && mProc.mRunning)
		{	mProc.terminate();
			mProc=null;
		}
	}
	
	private static class IntrayProcessor extends Thread implements Runnable
	{
	    public boolean mRunning=true;
	    private Configuration mConfig;
	    private IntrayTransport mTransport;
		
		public IntrayProcessor(INKFRequestContext context, IntrayTransport aTransport)
		{	mTransport=aTransport;
			try
			{	mConfig=configure(context);  //Must configure first time with the postCommission context
				context.logFormatted(INKFRequestContext.LEVEL_INFO, "MSG_INTRAY_START", mConfig.mIntray.toURI().toString());
				this.start();
			}
			catch(Exception e)
			{	context.logRaw(INKFRequestContext.LEVEL_WARNING, e.toString());				
			}
		}

	    public void terminate()
	    {	mRunning = false;
	    	mTransport.getTransportContext().logFormatted(INKFRequestContext.LEVEL_INFO, "MSG_INTRAY_STOP", mConfig.mIntray.toURI().toString());
	    	interrupt();
	    	mConfig=null;
	    }
	    
	    private Configuration configure(INKFRequestContext context) throws Exception
		{	Configuration cfg=new Configuration();
			cfg.mConfigurationResponse=context.sourceForResponse("param:config", Document.class);
			XMLReadable xr=new XMLReadable(cfg.mConfigurationResponse.getRepresentation());
			String s=xr.getTrimText("/config/in");			
			cfg.mIntray=new File(URI.create(s));
			if(!cfg.mIntray.canRead())
			{	throw new Exception("Intray Tranport: "+s+" is not readable by the current user");			
			}
			if(!cfg.mIntray.isDirectory())
			{	throw new Exception("Intray Tranport: "+s+" is not a directory");			
			}
			cfg.mInExt = xr.getTrimText("/config/in/@ext");	
			String regex = xr.getTrimText("/config/in/@regex");
			if(!cfg.mInExt.equals("") && !regex.equals(""))
			{	context.logRaw(context.LEVEL_WARNING, "@regex and @ext specified on <in>, @regex takes precedence");				
			}
			s=xr.getTrimText("/config/out");
			cfg.mOuttray=new File(URI.create(s));
			if(!cfg.mOuttray.canWrite())
			{	throw new Exception("Intray Tranport: "+s+" is not writeable by the current user");			
			}
			if(!cfg.mOuttray.isDirectory())
			{	throw new Exception("Intray Tranport: "+s+" is not a directory");			
			}
			cfg.mOutExt = xr.getTrimText("/config/out/@ext");
			String rewrite = xr.getTrimText("/config/out/@rewrite");
			if(!cfg.mOutExt.equals("") && !rewrite.equals(""))
			{	context.logRaw(context.LEVEL_WARNING, "@rewrite and @ext specified on <out>, @rewrite takes precedence");				
			}
			if(!regex.equals(""))
			{	if(rewrite.equals(""))
				{	context.logRaw(context.LEVEL_WARNING, "@regex specified but no @rewrite on <out>, please fix this");
					throw new Exception("Incomplete regex rewrite on out box problem.  @regex specified but need @rewrite on <out>");
				}
				cfg.mRegexRule=new RegexRewriterRule(regex, rewrite);				
			}
			if(regex.equals("") && !rewrite.equals(""))
			{	context.logRaw(context.LEVEL_WARNING, "@rewrite specified but no @regex on <in>, please fix this");
				throw new Exception("@rewrite specified but no @regex on <in>");
			}
			
			s=xr.getTrimText("/config/poll");
			cfg.mPollPeriod=Integer.parseInt(s);
			Node req=xr.getFirstNode("/config/request");
			cfg.mBuilder=new RequestBuilder(req, context.getKernelContext().getKernel().getLogger());
			context.logFormatted(INKFRequestContext.LEVEL_INFO, "MSG_INTRAY_CONFIG_LOADED", cfg.mIntray.toURI().toString());
			return cfg;
		}

	    public void run()
	    {   while (mRunning)
	    	{	if(mConfig.mConfigurationResponse.isExpired())
	        	{	try
	        		{	mConfig=configure(mTransport.getTransportContext());
	        		}
	        		catch(Exception e)
	        		{	mTransport.getTransportContext().logRaw(INKFRequestContext.LEVEL_WARNING, e.toString());	        			
	        		}
	        	}
	            File infl[] = mConfig.mIntray.listFiles();
				/** Sort files by last modification */
	            Arrays.sort(infl, new Comparator<File>() {
	                public int compare(File f1, File f2) {
	                    if (f1.lastModified() <f2.lastModified())
	                    {	return -1;
	                    }
	                    else if (f1.lastModified() > f2.lastModified())
	                    {	return +1;
	                    } else
	                    {	return 0;
	                    }
	                }
	            });
	            for(File f : infl)
	            {
	            	String furi=f.toURI().toString();           	
	            	String fnameWithExt = f.getName();
	            	RequestBuilder.Arguments args=new RequestBuilder.Arguments();
	            	args.addArgument("arg:file", furi);
	            	if(mConfig.mRegexRule!=null)
	            	{	if(!mConfig.mRegexRule.matches(fnameWithExt))
	            		{	continue; // skip (they specified an input file regex and this doesn't match!)
	            		}
	            	}
	            	else
	            	{
						String ext = "";
						String fname = "";					
						int extPos = fnameWithExt.lastIndexOf('.');
						if(extPos > 0 && extPos < fnameWithExt.length()) {
							fname = fnameWithExt.substring(0, extPos);
							ext = fnameWithExt.substring(extPos + 1);	            		
						}	            	
		            	if(mConfig.mInExt != null && !mConfig.mInExt.equals("") && !mConfig.mInExt.equals(ext)) {
		            		continue; // skip (they specified an input file type extension and this ain't it!)
		            	}
		           		args.addArgument("arg:filename", fname);
		           		args.addArgument("arg:extension", ext);
	            	}
	         
	           		INKFRequestContext context=mTransport.getTransportContext();
	            	try
	            	{	context.logFormatted(INKFRequestContext.LEVEL_INFO, "MSG_PROCESS_FILE", furi);	   
		            	INKFRequest req=mConfig.mBuilder.buildRequest(context, args, getContextClassLoader());	
		            	req.setRepresentationClass(IBinaryStreamRepresentation.class);
		            	IBinaryStreamRepresentation result=(IBinaryStreamRepresentation)context.issueRequest(req);

		            	String ouri=null;
		            	if(mConfig.mRegexRule!=null)
		            	{	String f2=mConfig.mRegexRule.map(fnameWithExt);
		            		ouri=mConfig.mOuttray.toURI().toString()+f2;
		            	}
		            	else
		            	{
			            	ouri=mConfig.mOuttray.toURI().toString()+f.getName();
			            	if(mConfig.mOutExt != null && !"".equals(mConfig.mOutExt)) {
			            		int extPos = ouri.lastIndexOf('.');
			            		if(extPos > 0) {
			            			ouri = ouri.substring(0, extPos);
			            			if(!mConfig.mOutExt.startsWith(".")) {
			            				ouri += "."; 
			            			}
			            			ouri += mConfig.mOutExt;
			            		}
			            	}
		            	}
		            	context.sink(ouri,result);
		            	context.logFormatted(INKFRequestContext.LEVEL_INFO, "MSG_PROCESS_FILE_COMPLETE", ouri);
		            	f.delete();
	            	}
	            	catch(Exception e)
	            	{	context.logRaw(INKFRequestContext.LEVEL_WARNING, e.toString());
	            	}
	            	if(!mRunning)
	            	{	break;
	            	}
	            }
	            try
	            {	sleep(mConfig.mPollPeriod*1000);
	            } catch (InterruptedException e)
	            {	//We've been interrupted   
	            }
	        }
	    }
	    
	    private class Configuration
		{	public INKFResponseReadOnly<Document> mConfigurationResponse;
			public File mIntray;
			public String mInExt;
			public File mOuttray;
			public String mOutExt;
			public int mPollPeriod;
			public RequestBuilder mBuilder;
			public RegexRewriterRule mRegexRule;
		}
	}
	
	
}